;;;; tcc.scm

(declare
  (fixnum)
  (hide error-list set-error-func tcc:error check)
  (foreign-declare #<<EOF
#include <libtcc.h>
EOF
) )

(define-foreign-type tcc-state* (nonnull-pointer "TCCState"))

(define-external (error_callback (c-pointer opaque) (c-string msg)) void
  ;(print msg)
  (error-list (cons msg (error-list))) )

(define tcc:new
  (let ([new (foreign-lambda tcc-state* "tcc_new")])
    (lambda ()
      (let ([s (new)])
	(set-error-func s #f (location error_callback))
	s) ) ) )

(define (tcc:error loc msg args)
  (signal
   (make-composite-condition 
    (make-property-condition 'exn 'message msg 'arguments args 'location loc)
    (make-property-condition 'tcc) ) ) )

(define (check loc thunk msg)
  (parameterize ([error-list '()])
    (let* ([r (thunk)]
	   [el (error-list)] )
      (if (null? el)
	  r
	  (tcc:error loc msg el) ) ) ) )

(define tcc:delete (foreign-lambda void "tcc_delete" tcc-state*))
(define tcc:enable-debug (foreign-lambda void "tcc_enable_debug" tcc-state*))

(define error-list (make-parameter '()))

(define set-error-func (foreign-lambda void "tcc_set_error_func" tcc-state* c-pointer nonnull-c-pointer))

(define tcc:set-warning (foreign-lambda bool "tcc_set_warning" tcc-state* c-string bool))
(define tcc:add-include-path (foreign-lambda void "tcc_add_include_path" tcc-state* c-string))
(define tcc:add-sysinclude-path (foreign-lambda void "tcc_add_sysinclude_path" tcc-state* c-string))
(define tcc:define-symbol (foreign-lambda void "tcc_define_symbol" tcc-state* c-string c-string))
(define tcc:undefine-symbol (foreign-lambda void "tcc_undefine_symbol" tcc-state* c-string))
(define tcc:add-file (foreign-lambda int "tcc_add_file" tcc-state* c-string))

(define tcc:compile-string
  (let ([compile (foreign-safe-lambda int "tcc_compile_string" tcc-state* c-string)])
    (lambda (s str)
      (check 'tcc:compile-string (cut compile s str) "errors during compilation") ) ) )

(define-foreign-variable tcc_output_memory int "TCC_OUTPUT_MEMORY")
(define-foreign-variable tcc_output_exe int "TCC_OUTPUT_EXE")
(define-foreign-variable tcc_output_dll int "TCC_OUTPUT_DLL")
(define-foreign-variable tcc_output_obj int "TCC_OUTPUT_OBJ")
(define tcc/output-memory tcc_output_memory)
(define tcc/output-exe tcc_output_exe)
(define tcc/output-dll tcc_output_dll)
(define tcc/output-obj tcc_output_obj)

(define tcc:set-output-type (foreign-lambda void "tcc_set_output_type" tcc-state* int))
(define tcc:add-library-path (foreign-lambda void "tcc_add_library_path" tcc-state* c-string))
(define tcc:add-library (foreign-lambda void "tcc_add_library" tcc-state* c-string))

(define tcc:add-symbol
  (let ([add (foreign-lambda void "tcc_add_symbol" tcc-state* c-string c-pointer)])
    (lambda (s str ptr)
      (check 'tcc:add-symbol (cut add s str ptr) "could not add symbol") ) ) )

(define tcc:output-file
  (let ([output (foreign-lambda int "tcc_output_file" tcc-state* c-string)])
    (lambda (s fname)
      (check 'tcc:output-file (cut output s fname) "generation of output file failed") ) ) )

(define tcc:run
  (foreign-safe-lambda* int ([tcc-state* state] [scheme-object args])
    "char **argv, *ptr;"
    "C_word arg;"
    "int i, n, argc = C_unfix(C_i_length(args));"
    "argv = (char **)C_malloc(argc * sizeof(char *));"
    "if(argv == NULL) return(999);"
    "for(i = 0; i < argc; ++i) {"
    "  arg = C_block_item(args, 0);"
    "  ptr = (char *)C_malloc((n = C_header_size(arg)) + 1);"
    "  if(ptr == NULL) return(998);"
    "  C_memcpy(ptr, C_data_pointer(arg), n);"
    "  ptr[ n ] = '\\0';"
    "  argv[ i ] = ptr;"
    "  args = C_block_item(args, 1);"
    "}"
    "n = tcc_run(state, argc, argv);"
    "for(i = 0; i < argc; ++i) C_free(argv[ i ]);"
    "C_free(argv);"
    "return(n);") )

(define tcc:relocate
  (let ([rel (foreign-lambda int "tcc_relocate" tcc-state*)])
    (lambda (s)
      (unless (zero? (rel s))
	(tcc:error 'tcc:relocate "relocation failed" '()) ) ) ) )

(define tcc:get-symbol
  (let ([get (foreign-lambda int "tcc_get_symbol" tcc-state* c-pointer c-string)])
    (lambda (s str)
      (let-location ([val c-pointer])
        (if (zero? (get s (location val) str))
            val
            (tcc:error 'tcc:get-symbol "can't get symbol" '()))))))
